/**
 * My Nixie Clock IN-12
 * Vladimir N. Shilov <shilow@ukr.net>
 * 2019.01.18
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "main.h"
#include "i2c.h"
#include "ds3231.h"
#include "rtos.h"
#include "event-system.h"

/* Defines */
#define TIMER2_HZ		      250
#define TIMER2_PRESCALER  256
#define TIMER2_CS         (1<<CS22 | 1<<CS21 | 0<<CS20)
#define	TIMER2_CNT		    (0x100 - (F_CPU / TIMER2_PRESCALER / TIMER2_HZ))

/* Variables */
static volatile uint8_t Digit[4] = {1, 2, 3, 4};
static rtc_t RTC;
volatile struct {
	uint8_t RTC_Int:	1;
	uint8_t rezerv:   7;
} Flag;

/* Constants */
/* Function prototypes */
void Board_Init(void);

void main(void) {
  /**
   * Локальные переменные
   */
  uint8_t event = 0;
  Flag.RTC_Int = 0;

  /**
   * Инициализация, настройка...
   */
  Board_Init();

  // Initialize Scheduler
  RTOS_Init();

  // Initialize I2C Bus and RTC
  I2C_Init();
  RTC_Init();
  RTC_ReadAll(&RTC);

  // Initialize Event State Machine
  ES_Init(stNoChange);

  /** main loop */
  do {

    if (Flag.RTC_Int != 0) {
      Flag.RTC_Int = 0;
      RTC_ReadTime(&RTC);
      Digit[0] = RTC.Min >> 4;
      Digit[1] = RTC.Min & 0x0F;
      Digit[2] = RTC.Sec >> 4;
      Digit[3] = RTC.Sec & 0x0F;
    }
    // крутим диспетчер
    RTOS_DispatchTask();

    event = ES_GetEvent();
    if (event) {
      ES_Dispatch(event);
    }

  } while(1);
}

/**
 *  П о д п р о г р а м м ы
 */

/**
 * @brief Initializy perephireal
 */
void Board_Init(void) {
  /* power off Analog Comparator */
  ACSR = ACD;

  /* GPIO */
  PORTB = BUTTON_PINS; // enable pull-up
  DDRC = DIGIT_PINS; // as output
  DDRD = (DOT_PIN | ANOD_PINS); // as output

  /* Timer2 - refresh Nixie values */
  TCCR2 = TIMER2_CS;
  TCNT2 = TIMER2_CNT;
  TIMSK = _BV(TOIE2);

  /* Interrupt from RTC */
  MCUCR = _BV(ISC11); // falling edge
  GICR = _BV(INT1);

  /* Enable Interrupts */
  sei();
}

/**
 *  П р е р ы в а н и я
 */

/**
 * @brief RTC one seconds interrupt
 */
ISR (INT0_vect) {
	// Blink Dot for test
	PIND |= DOT_PIN;
	Flag.RTC_Int = 1;
}

/**
 * @brief Refresh Nixie output
 * @note Digit[] must be in range 0x00 - 0x0F
 */
ISR(TIMER2_OVF_vect) {
  static uint8_t idx = 0;

  // reload timer
  TCNT2 = TIMER2_CNT;

  // read current register value and clean bits
  uint8_t pd = PORTD & ~ANOD_PINS;
  uint8_t pc = PORTC & ~DIGIT_PINS;

  // power off lamps
  PORTD = pd;
  PORTC =  pc;

  switch (idx) {
  case 0:
    // output lamp value
    PORTC = pc | Digit[0];
    // power on lamp
    PORTD = pd | ANOD1;
    idx = 1;
    break;

  case 1:
    PORTC = pc | Digit[1];
    PORTD = pd | ANOD2;
    idx = 2;
    break;

  case 2:
    PORTC = pc | Digit[2];
    PORTD = pd | ANOD3;
    idx = 3;
    break;

  case 3:
    PORTC = pc | Digit[3];
    PORTD = pd | ANOD4;
    idx = 0;
    break;

  default:
    idx = 0;
    break;
  }
}

/**
 * @brief заглушка для неиспользуемых прерываний
 */
ISR(__vector_default,ISR_NAKED) {
	reti();
}
